import asyncio
import tempfile
import time
from threading import Thread

import dagster_pandas as dagster_pd
import pytest
from dagster_dask import DataFrame, dask_executor
from dask.distributed import Scheduler, Worker

from dagster import (
    DagsterUnmetExecutorRequirementsError,
    InputDefinition,
    ModeDefinition,
    VersionStrategy,
    execute_pipeline,
    execute_pipeline_iterator,
    file_relative_path,
    fs_io_manager,
    job,
    op,
    pipeline,
    reconstructable,
    solid,
)
from dagster.core.definitions.executor_definition import default_executors
from dagster.core.definitions.reconstruct import ReconstructablePipeline
from dagster.core.events import DagsterEventType
from dagster.core.test_utils import instance_for_test, nesting_composite_pipeline
from dagster.utils import send_interrupt


@solid
def simple(_):
    return 1


@pipeline(
    mode_defs=[
        ModeDefinition(
            name="default",
            executor_defs=default_executors + [dask_executor],
        ),
        ModeDefinition(
            name="filesystem",
            resource_defs={"io_manager": fs_io_manager},
            executor_defs=default_executors + [dask_executor],
        ),
    ]
)
def dask_engine_pipeline():
    simple()


def test_execute_on_dask_local():
    with tempfile.TemporaryDirectory() as tempdir:
        with instance_for_test(temp_dir=tempdir) as instance:
            result = execute_pipeline(
                reconstructable(dask_engine_pipeline),
                run_config={
                    "resources": {"io_manager": {"config": {"base_dir": tempdir}}},
                    "execution": {"dask": {"config": {"cluster": {"local": {"timeout": 30}}}}},
                },
                instance=instance,
                mode="filesystem",
            )
            assert result.result_for_solid("simple").output_value() == 1


def dask_composite_pipeline():
    return nesting_composite_pipeline(
        6,
        2,
        mode_defs=[
            ModeDefinition(
                resource_defs={"io_manager": fs_io_manager},
                executor_defs=default_executors + [dask_executor],
            )
        ],
    )


def test_composite_execute():
    with instance_for_test() as instance:
        result = execute_pipeline(
            reconstructable(dask_composite_pipeline),
            run_config={
                "execution": {"dask": {"config": {"cluster": {"local": {"timeout": 30}}}}},
            },
            instance=instance,
        )
        assert result.success


@solid(input_defs=[InputDefinition("df", dagster_pd.DataFrame)])
def pandas_solid(_, df):  # pylint: disable=unused-argument
    pass


@pipeline(
    mode_defs=[
        ModeDefinition(
            resource_defs={"io_manager": fs_io_manager},
            executor_defs=default_executors + [dask_executor],
        )
    ]
)
def pandas_pipeline():
    pandas_solid()


def test_pandas_dask():
    run_config = {
        "solids": {
            "pandas_solid": {
                "inputs": {"df": {"csv": {"path": file_relative_path(__file__, "ex.csv")}}}
            }
        }
    }

    with instance_for_test() as instance:
        result = execute_pipeline(
            ReconstructablePipeline.for_file(__file__, pandas_pipeline.name),
            run_config={
                "execution": {"dask": {"config": {"cluster": {"local": {"timeout": 30}}}}},
                **run_config,
            },
            instance=instance,
        )

        assert result.success


@solid(input_defs=[InputDefinition("df", DataFrame)])
def dask_solid(_, df):  # pylint: disable=unused-argument
    pass


@pipeline(
    mode_defs=[
        ModeDefinition(
            resource_defs={"io_manager": fs_io_manager},
            executor_defs=default_executors + [dask_executor],
        )
    ]
)
def dask_pipeline():
    dask_solid()


def test_dask():
    run_config = {
        "solids": {
            "dask_solid": {
                "inputs": {
                    "df": {"read": {"csv": {"path": file_relative_path(__file__, "ex*.csv")}}}
                }
            }
        }
    }
    with instance_for_test() as instance:

        result = execute_pipeline(
            ReconstructablePipeline.for_file(__file__, dask_pipeline.name),
            run_config={
                "execution": {"dask": {"config": {"cluster": {"local": {"timeout": 30}}}}},
                **run_config,
            },
            instance=instance,
        )

    assert result.success


def test_execute_on_dask_local_without_io_manager():
    with pytest.raises(DagsterUnmetExecutorRequirementsError):
        with instance_for_test() as instance:
            result = execute_pipeline(
                reconstructable(dask_engine_pipeline),
                run_config={
                    "execution": {"dask": {"config": {"cluster": {"local": {"timeout": 30}}}}},
                },
                instance=instance,
                mode="default",
            )
            assert result.result_for_solid("simple").output_value() == 1


@solid(input_defs=[InputDefinition("df", DataFrame)])
def sleepy_dask_solid(_, df):  # pylint: disable=unused-argument
    start_time = time.time()
    while True:
        time.sleep(0.1)
        if time.time() - start_time > 120:
            raise Exception("Timed out")


@pipeline(
    mode_defs=[
        ModeDefinition(
            resource_defs={"io_manager": fs_io_manager},
            executor_defs=default_executors + [dask_executor],
        )
    ]
)
def sleepy_dask_pipeline():
    sleepy_dask_solid()


def test_dask_terminate():
    run_config = {
        "solids": {
            "sleepy_dask_solid": {
                "inputs": {
                    "df": {"read": {"csv": {"path": file_relative_path(__file__, "ex*.csv")}}}
                }
            }
        }
    }

    interrupt_thread = None
    result_types = []

    with instance_for_test() as instance:
        for result in execute_pipeline_iterator(
            pipeline=ReconstructablePipeline.for_file(__file__, sleepy_dask_pipeline.name),
            run_config=run_config,
            instance=instance,
        ):
            # Interrupt once the first step starts
            if result.event_type == DagsterEventType.STEP_START and not interrupt_thread:
                interrupt_thread = Thread(target=send_interrupt, args=())
                interrupt_thread.start()

            if result.event_type == DagsterEventType.STEP_FAILURE:
                assert (
                    "DagsterExecutionInterruptedError" in result.event_specific_data.error.message
                )

            result_types.append(result.event_type)

        interrupt_thread.join()

        assert DagsterEventType.STEP_FAILURE in result_types
        assert DagsterEventType.PIPELINE_FAILURE in result_types


@pytest.mark.skip(
    "Failing with RuntimeError: This event loop is already running since distributed==2022.1.0"
)
def test_existing_scheduler():
    def _execute(scheduler_address, instance):
        return execute_pipeline(
            reconstructable(dask_engine_pipeline),
            run_config={
                "execution": {
                    "dask": {"config": {"cluster": {"existing": {"address": scheduler_address}}}}
                },
            },
            instance=instance,
            mode="filesystem",
        )

    async def _run_test():
        with instance_for_test() as instance:
            async with Scheduler() as scheduler:
                async with Worker(scheduler.address) as _:
                    result = await asyncio.get_event_loop().run_in_executor(
                        None, _execute, scheduler.address, instance
                    )
                    assert result.success
                    assert result.result_for_solid("simple").output_value() == 1

    asyncio.get_event_loop().run_until_complete(_run_test())


@solid
def foo_solid():
    return "foo"


class BasicVersionStrategy(VersionStrategy):
    def get_solid_version(self, _):
        return "foo"


@pipeline(
    mode_defs=[
        ModeDefinition(
            resource_defs={"io_manager": fs_io_manager},
            executor_defs=default_executors + [dask_executor],
        )
    ],
    version_strategy=BasicVersionStrategy(),
)
def foo_pipeline():
    foo_solid()


def test_dask_executor_memoization():
    with instance_for_test() as instance:
        result = execute_pipeline(
            reconstructable(foo_pipeline),
            instance=instance,
            run_config={"execution": {"dask": {"config": {"cluster": {"local": {"timeout": 30}}}}}},
        )
        assert result.success
        assert result.output_for_solid("foo_solid") == "foo"

        result = execute_pipeline(
            reconstructable(foo_pipeline),
            instance=instance,
            run_config={"execution": {"dask": {"config": {"cluster": {"local": {"timeout": 30}}}}}},
        )
        assert result.success
        assert len(result.step_event_list) == 0


@op
def the_op():
    return 5


@job(executor_def=dask_executor)
def the_job():
    the_op()


def test_dask_executor_job():
    with instance_for_test() as instance:
        result = execute_pipeline(
            reconstructable(the_job),
            instance=instance,
            run_config={"execution": {"config": {"cluster": {"local": {"timeout": 30}}}}},
        )
        assert result.success
        assert result.output_for_solid("the_op") == 5



from dagster.core.utils import check_dagster_package_version

from .data_frame import DataFrame
from .executor import dask_executor
from .resources import dask_resource
from .version import __version__

check_dagster_package_version("dagster-dask", __version__)

__all__ = [
    "DataFrame",
    "dask_executor",
]



import dask.dataframe as dd
import pytest
from dagster_dask import DataFrame
from dagster_dask.data_frame import DataFrameReadTypes, DataFrameToTypes
from dagster_dask.utils import DataFrameUtilities
from dask.dataframe.utils import assert_eq

from dagster import InputDefinition, execute_solid, file_relative_path, solid


def create_dask_df():
    path = file_relative_path(__file__, "num.csv")
    return dd.read_csv(path)


@pytest.mark.parametrize(
    "file_type",
    [
        pytest.param("csv", id="csv"),
        pytest.param("parquet", id="parquet"),
        pytest.param("json", id="json"),
    ],
)
def test_dataframe_inputs(file_type):
    @solid(input_defs=[InputDefinition(dagster_type=DataFrame, name="input_df")])
    def return_df(_, input_df):
        return input_df

    file_name = file_relative_path(__file__, f"num.{file_type}")

    read_result = execute_solid(
        return_df,
        run_config={
            "solids": {
                "return_df": {"inputs": {"input_df": {"read": {file_type: {"path": file_name}}}}}
            }
        },
    )
    assert read_result.success
    assert assert_eq(read_result.output_value(), create_dask_df())


def test_dataframe_loader_config_keys_dont_overlap():
    """
    Test that the read_keys, which are deprecated, do not overlap with
    the normal loader config_keys.
    """
    config_keys = set(DataFrameUtilities.keys())
    config_keys.add("read")
    read_keys = set(DataFrameReadTypes.keys())

    assert len(config_keys.intersection(read_keys)) == 0


def test_dataframe_materializer_config_keys_dont_overlap():
    """
    Test that the to_keys, which are deprecated, do not overlap with
    the normal materializer config_keys.
    """
    config_keys = set(DataFrameUtilities.keys())
    config_keys.add("to")
    to_keys = set(DataFrameToTypes.keys())

    assert len(config_keys.intersection(to_keys)) == 0



import dask.dataframe as dd
from dagster_dask import DataFrame

from dagster import InputDefinition, execute_solid, file_relative_path, solid


@solid(input_defs=[InputDefinition(dagster_type=DataFrame, name="input_df")])
def passthrough(_, input_df: DataFrame) -> DataFrame:
    return input_df


def generate_config(path, **df_opts):
    return {
        "solids": {
            "passthrough": {
                "inputs": {
                    "input_df": {
                        "read": {
                            "csv": {
                                "path": path,
                            },
                        },
                        **df_opts,
                    },
                },
            },
        },
    }


def test_drop():
    path = file_relative_path(__file__, "canada.csv")
    col = "current"

    input_df = dd.read_csv(path)
    assert col in input_df.columns

    # Drop the "current" column.
    run_config = generate_config(path, drop={"columns": col})
    result = execute_solid(passthrough, run_config=run_config)
    output_df = result.output_value()
    assert col not in output_df.columns


def test_sample():
    path = file_relative_path(__file__, "canada.csv")
    frac = 0.25

    input_df = dd.read_csv(path)

    run_config = generate_config(path, sample={"frac": frac})
    result = execute_solid(passthrough, run_config=run_config)
    output_df = result.output_value()

    assert len(input_df) * frac == len(output_df)


def test_reset_index():
    path = file_relative_path(__file__, "canada.csv")

    input_df = dd.read_csv(path)

    # Reset the index without dropping. We expect the index to be moved
    # to the columns, as a column named "index".
    run_config = generate_config(path, reset_index={})
    result = execute_solid(passthrough, run_config=run_config)
    output_df = result.output_value()
    assert "index" in output_df.columns
    assert len(output_df.columns) == len(input_df.columns) + 1


def test_set_index():
    path = file_relative_path(__file__, "canada.csv")
    col = "ID"

    input_df = dd.read_csv(path)
    assert col in input_df.columns

    # Set index to ID. We expect the column to be dropped by default.
    run_config = generate_config(path, set_index={"other": col})
    result = execute_solid(passthrough, run_config=run_config)
    output_df = result.output_value()
    assert col not in output_df.columns

    # Set index to ID without dropping the column.
    run_config = generate_config(path, set_index={"other": col, "drop": False})
    result = execute_solid(passthrough, run_config=run_config)
    output_df = result.output_value()
    assert col in output_df.columns


def test_repartition():
    path = file_relative_path(__file__, "canada.csv")
    npartitions = 2

    input_df = dd.read_csv(path)
    assert input_df.npartitions == 1

    # Repartition from 1 to 2 partitions.
    run_config = generate_config(path, repartition={"npartitions": npartitions})
    result = execute_solid(passthrough, run_config=run_config)
    output_df = result.output_value()
    assert output_df.npartitions == npartitions


def test_normalize_column_names():
    path = file_relative_path(__file__, "canada.csv")

    input_df = dd.read_csv(path)
    assert all(col in input_df.columns for col in ("ID", "provinceOrTerritory", "country"))

    # Set normalize_column_names=False to not modify the column names
    run_config = generate_config(path, normalize_column_names=False)
    result = execute_solid(passthrough, run_config=run_config)
    output_df = result.output_value()
    assert all(col in output_df.columns for col in ("ID", "provinceOrTerritory", "country"))

    # Set normalize_column_names=True to modify the column names
    run_config = generate_config(path, normalize_column_names=True)
    result = execute_solid(passthrough, run_config=run_config)
    output_df = result.output_value()
    assert all(col in output_df.columns for col in ("id", "province_or_territory", "country"))


def test_utilities_combo():
    path = file_relative_path(__file__, "canada.csv")

    input_df = dd.read_csv(path)
    assert input_df.npartitions == 1

    # Apply multiple utilities at once. The utilities are expected to
    # apply in the following order, regardless of config order:
    #   sample, reset_index, set_index, repartition, normalize_column_names
    run_config = generate_config(
        path,
        normalize_column_names=True,
        set_index={"other": "ID", "drop": True},
        repartition={"npartitions": 3},
        sample={"frac": 0.5},
        reset_index={"drop": True},
    )
    result = execute_solid(passthrough, run_config=run_config)
    output_df = result.output_value()

    # sample(frac=0.5)
    assert len(input_df) * 0.5 == len(output_df)

    # reset_index(drop=True)
    assert "index" not in input_df.columns
    assert "index" not in output_df.columns

    # set_index(other="ID", drop=True)
    assert "ID" in input_df.columns
    assert "ID" not in output_df.columns

    # repartition(npartitions=3)
    assert input_df.npartitions == 1
    assert output_df.npartitions == 3

    # normalize_column_names(true)
    # No id due to it being set to the index and dropped.
    assert all(col in input_df.columns for col in ("ID", "provinceOrTerritory", "country"))
    assert all(col in output_df.columns for col in ("province_or_territory", "country"))
